@@ -1,16 +1,10 @@
-/*-------------------------------------------------------------------------
- *
- * pg_tde_fe.h
- *	   Frontened definitions for encrypted XLog storage manager
- *
- *-------------------------------------------------------------------------
+/*
+ * Frontened definitions for encrypted XLog storage manager
  */
 
 #ifndef PG_TDE_FE_INIT_H
 #define PG_TDE_FE_INIT_H
 
-#ifdef PERCONA_EXT
-
 #include "common/pg_tde_utils.h"
 #include "encryption/enc_aes.h"
 #include "keyring/keyring_file.h"
@@ -28,6 +22,4 @@ pg_tde_fe_init(const char *kring_dir)
 	pg_tde_set_data_dir(kring_dir);
 }
 
-#endif							/* PERCONA_EXT */
-
 #endif							/* PG_TDE_FE_INIT_H */
